# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import path as upath
from suanpan.app import app
from suanpan.app.modules.base import Module
from suanpan.storage import storage

module = Module()


@module.stream("storage.create")
def createStorage(context):
    args = context.args
    stream = context.stream
    filepath = storage.getPathInTempStore(args.key)
    stream.pipe(filepath)
    if stream.done:
        symlink(args.key)


@module.on("storage.remove")
def removeStrorage(context):
    args = context.args
    filepath = storage.getPathInTempStore(args.key)
    debugpath = app.sio.static("debug", args.key)
    storage.remove(args.key)
    upath.remove(filepath)
    upath.remove(debugpath)


@module.on("storage.list")
def listStorage(context):
    args = context.args
    return [meta(key) for key in storage.listFiles(args.key)]


@module.on("storage.meta.get")
def getStorageMeta(context):
    args = context.args
    return meta(args.key)


@module.fn
def notifyStorageMeta(key):
    app.sio.emit("storage.meta.notify", meta(key))


@module.fn
def meta(key):
    return {"src": f"debug/{key}", "key": key}


@module.fn
def dowload(key):
    filepath = storage.getPathInTempStore(key)
    storage.dowload(key, filepath)
    symlink(key)
    return filepath


@module.fn
def remove(key):
    filepath = storage.getPathInTempStore(key)
    debugpath = app.sio.static("debug", key)
    storage.remove(key)
    upath.remove(filepath)
    upath.remove(debugpath)
    return key


@module.fn
def symlink(key):
    filepath = storage.getPathInTempStore(key)
    debugpath = app.sio.static("debug", key)
    return upath.symlink(filepath, debugpath)



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.error import base


class NodeError(base.Error):
    MESSAGE = "Suanpan Node Error: {}"


class NodeAutoLoadError(NodeError):
    MESSAGE = "Suanpan Node Auto Load Error: {}"


class NodeAutoLoadUnsupportedDataSubtypeError(NodeError):
    MESSAGE = "Suanpan Node Auto Load Unsupported Data Subtype: {}"


class NodeAutoLoadUnsupportedParamTypeError(NodeError):
    MESSAGE = "Suanpan Node Auto Load Unsupported Param Type: {}"



# coding=utf-8
from __future__ import absolute_import, division, print_function

import os

from suanpan import path as spath
from suanpan import runtime
from suanpan.log import logger
from suanpan.storage import base


class Storage(base.Storage):
    def __init__(
        self,
        localTempStore=base.Storage.DEFAULT_TEMP_DATA_STORE,
        localGlobalStore=base.Storage.DEFAULT_GLOBAL_DATA_STORE,
        **kwargs,
    ):
        super(Storage, self).__init__(
            delimiter=os.sep,
            tempStore=localTempStore,
            globalStore=localGlobalStore,
            **kwargs,
        )

    def download(self, name, path):  # pylint: disable=unused-argument
        tmpPath = self.getPathInTempStore(name)
        if path == tmpPath:
            logger.info(
                f"Downloaded: {self.storageUrl(name)} -> {path} - ({self.name}) Nothing to do!"
            )
        else:
            logger.debug(
                f"Downloading: {self.storageUrl(name)} -> {path} - ({self.name})"
            )
            spath.copy(tmpPath, path)
            logger.info(
                f"Downloaded: {self.storageUrl(name)} -> {path} - ({self.name})"
            )
        return path

    def upload(self, name, path):  # pylint: disable=unused-argument
        tmpPath = self.getPathInTempStore(name)
        if path == tmpPath:
            logger.info(
                f"Uploaded: {self.storageUrl(name)} -> {path} - ({self.name}) Nothing to do!"
            )
        else:
            logger.debug(
                f"Uploading: {self.storageUrl(name)} -> {path} - ({self.name})"
            )
            spath.copy(path, tmpPath)
            logger.info(f"Uploaded: {self.storageUrl(name)} -> {path} - ({self.name})")
        return tmpPath

    def copy(self, path, dist):  # pylint: disable=unused-argument
        pathUrl = self.storageUrl(path)
        distUrl = self.storageUrl(dist)
        logger.debug(f"Copying: {pathUrl} -> {distUrl}")
        _path = self.getPathInTempStore(path)
        _dist = self.getPathInTempStore(dist)
        spath.copy(_path, _dist)
        logger.info(f"Copied: {pathUrl} -> {distUrl}")
        return dist

    def remove(self, path):  # pylint: disable=unused-argument
        logger.debug(f"Removing: {self.storageUrl(path)}")
        spath.remove(self.getPathInTempStore(path))
        logger.info(f"Removed: {self.storageUrl(path)}")
        return path

    def walk(self, folderName):
        root = self.getPathInTempStore(folderName)
        return runtime.saferun(os.walk, default=iter(()))(root)

    def listAll(self, folderName):
        root = self.getPathInTempStore(folderName)
        return (self.storagePathJoin(root, p) for p in os.listdir(root))

    def listFolders(self, folderName):
        return (p for p in self.listAll(folderName) if os.path.isdir(p))

    def listFiles(self, folderName):
        return (p for p in self.listAll(folderName) if os.path.isfile(p))

    def isFolder(self, folderName):
        folder = self.getPathInTempStore(folderName)
        return os.path.isdir(folder)

    def isFile(self, objectName):
        file = self.getPathInTempStore(objectName)
        return os.path.isfile(file)

    def storagePathJoin(self, *paths):
        return self.localPathJoin(*paths)

    def storageRelativePath(self, path, basepath):
        return self.localRelativePath(path, basepath)

    def storageUrl(self, path):
        return "file://" + self.getPathInTempStore(path)

    def getStorageMd5(self, objectName):
        return self.getLocalMd5(self.getPathInTempStore(objectName))

    def getStorageSize(self, objectName):
        return self.getLocalSize(self.getPathInTempStore(objectName))



# coding=utf-8
from __future__ import absolute_import, print_function

import contextlib
import copy
import itertools
from collections import defaultdict

from suanpan import error, runtime
from suanpan.arguments.auto import AutoArg
from suanpan.interfaces import (
    HasArguments,
    HasCallHooks,
    HasCleanHooks,
    HasExitHooks,
    HasInitHooks,
    HasSaveHooks,
)
from suanpan.log import logger
from suanpan.objects import Context


class Arguments(Context):
    @classmethod
    def froms(cls, *args):
        context = cls()
        context.update({arg.key: arg.value for arg in args if arg.key})
        context.update({arg.alias: arg.value for arg in args if arg.alias})
        return context


class Result(Context):
    pass


class Component(
    HasArguments, HasInitHooks, HasCallHooks, HasSaveHooks, HasCleanHooks, HasExitHooks
):
    def __init__(self):
        super(Component, self).__init__()
        self.runFunc = None
        self.arguments = defaultdict(list)
        self.argsDict = {}

    def __call__(self, *arg, **kwargs):
        return self.run(*arg, **kwargs)

    def start(self, *arg, **kwargs):
        return self.run(*arg, **kwargs)

    @property
    def name(self):
        return self.runFunc.__name__

    @runtime.globalrun
    def run(self, *arg, **kwargs):
        self.callBeforeInitHooks()
        context = self.init(*arg, **kwargs)
        self.callAfterInitHooks(context)
        self.registerBeforeExitHooks(context)

        self.callBeforeCallHooks(context)
        if not self.runFunc:
            raise error.ComponentHandlerNotSet()
        if not callable(self.runFunc):
            raise error.ComponentHandlerNotCallable(self.name)
        results = self.runFunc(context)
        self.callAfterCallHooks(context)

        self.callBeforeSaveHooks(context)
        outputs = self.save(results)
        self.callAfterSaveHooks(context)

        self.callBeforeCleanHooks(context)
        self.clean()
        self.callAfterCleanHooks()

        return outputs

    def init(self, *args, **kwargs):
        self.argsDict = self.getArgsDict(*args, **kwargs)
        globalArgs = self.loadGlobalArguments(self.argsDict)
        self.initBase(globalArgs)
        context = self._getContext(globalArgs)
        args = self.loadComponentArguments(self.argsDict)
        context.update(args=args)
        return context

    def load(self, args, argsDict=None):
        logger.debug("Loading Arguments:")
        argsDict = argsDict or self.argsDict
        for arg in args:
            if isinstance(arg, AutoArg):
                arg.setBackend(argtype="inputs")
        args = self.loadFormatArguments(args, argsDict)
        return Arguments.froms(*args)

    def loadGlobalArguments(self, argsDict=None):
        logger.debug("Loading Global Arguments:")
        argsDict = argsDict or self.argsDict
        args = self.getGlobalArguments()
        args = self.loadFormatArguments(args, argsDict)
        return Arguments.froms(*args)

    def loadComponentArguments(self, argsDict=None):
        logger.debug("Loading Component Arguments:")
        argsDict = argsDict or self.argsDict
        args = copy.deepcopy(self.getArguments(exclude="outputs"))
        args = self.loadFormatArguments(args, argsDict)
        return Arguments.froms(*args)

    def initBase(self, args):
        pass

    @contextlib.contextmanager
    def context(self, args=None):  # pylint: disable=unused-argument
        yield Context()

    def _getContext(self, *args, **kwargs):
        self.contextManager = self.context(  # pylint: disable=attribute-defined-outside-init
            *args, **kwargs
        )
        return next(self.contextManager.gen)  # pylint: disable-msg=e1101

    def _closeContext(self):
        try:
            next(self.contextManager.gen)  # pylint: disable-msg=e1101
        except StopIteration:
            pass

    def getArguments(self, include=None, exclude=None):
        includes = set(self.arguments.keys() if not include else self._list(include))
        excludes = set([] if not exclude else self._list(exclude))
        includes = includes - excludes
        argumentsLists = [self.arguments[c] for c in includes]
        return list(itertools.chain(*argumentsLists))

    def save(self, results, args=None, **_):
        logger.debug("Saving...")
        if args:
            for arg in args:
                if isinstance(arg, AutoArg):
                    arg.setBackend(argtype="outputs")
        else:
            args = copy.deepcopy(self.getArguments(include="outputs"))
        args = self.loadCleanArguments(args, self.argsDict)
        return self.saveOutputs(args, results)

    def saveOutputs(self, args, results):
        if len(args) > 1:
            return self.saveMutipleOutputs(args, results)
        if len(args) == 1:
            return self.saveOneOutput(args[0], results)
        return None

    def saveMutipleOutputs(self, args, results):
        if isinstance(results, (tuple, list)):
            results = (Result.froms(value=result) for result in results)
            return self.saveArguments(args, results)
        if isinstance(results, dict):
            tmp = [
                (
                    argument,
                    Result.froms(
                        value=self.getArgumentValueFromDict(results, argument)
                    ),
                )
                for argument in args
                if self.getArgumentValueFromDict(results, argument) is not None
            ]
            if not tmp:
                return self.saveOneOutput(args[0], results)
            args, results = zip(*tmp)
            return self.saveArguments(args, results)
        raise error.ComponentError(f"Incorrect results {results}")

    def saveOneOutput(self, arg, results):
        result = (
            Result.froms(value=self.getArgumentValueFromDict(results, arg))
            if isinstance(results, dict) and self.hasArgumentValueFromDict(results, arg)
            else Result.froms(value=results)
        )
        return {arg.key: arg.save(result)}

    def clean(self):
        self._closeContext()

    def addArgument(self, arg, argtype="args", reverse=False):
        if reverse:
            self.arguments[argtype].insert(0, arg)
        else:
            self.arguments[argtype].append(arg)

    def use(self, func):
        self.runFunc = func
        return self

    def arg(self, argument, argtype="args", reverse=False):
        if isinstance(argument, AutoArg):
            argument.setBackend(argtype=argtype)
        self.addArgument(argument, argtype=argtype, reverse=reverse)
        return lambda *args, **kwargs: self

    def input(self, *args, **kwargs):
        kwargs.update(argtype="inputs")
        return self.arg(*args, **kwargs)

    def output(self, *args, **kwargs):
        kwargs.update(argtype="outputs")
        return self.arg(*args, **kwargs)

    def param(self, *args, **kwargs):
        kwargs.update(argtype="params")
        return self.arg(*args, **kwargs)

    def column(self, *args, **kwargs):
        kwargs.update(argtype="columns")
        return self.arg(*args, **kwargs)

    def _list(self, params=None):
        return [params] if isinstance(params, str) else list(params)



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import api, error, g, objects


class Graph(object):
    def __init__(self, appId=g.appId):
        self.appId = appId
        self.graph = self.load(self.appId)

    def load(self, appId):
        return objects.Context(api.app.getAppGraph(appId))

    def reload(self):
        self.graph = self.load(self.appId)

    def update(self):
        api.app.updateAppGraph(self.appId, self.graph)

    def revert(self):
        api.app.revertAppGraph(self.appId)

    def getNodeById(self, nodeId):
        node = self.graph.processes[nodeId]
        if not node:
            raise error.GraphError(f"No such node {nodeId}")
        node.id = nodeId
        return node

    def getNodeByName(self, name):
        for nodeId, node in self.graph.processes.items():
            if node.metadata.label == name:
                node.id = nodeId
                return node
        raise error.GraphError(f"No such node {name}")

    def addConnection(self, sourceNodeId, sourcePortId, targetNodeId, targetPortId):
        self.graph.connections.append(
            {
                "src": {"process": sourceNodeId, "port": sourcePortId},
                "tgt": {"process": targetNodeId, "port": targetPortId},
            }
        )
        return self

    def modGlobalVar(self, globalVariables):
        for i, var in enumerate(globalVariables):
            self.graph.properties.globalVars[i].update(var)
        return self

    def modNodeParam(self, nodeId, params):
        getattr(self.graph.processes, nodeId).metadata["def"].params.update(params)
        return self



# coding=utf-8
from __future__ import absolute_import, print_function

import base64
import hashlib
import hmac

import requests

from suanpan import error, g, utils
from suanpan.log import logger


def signatureV1(secret, data):
    h = hmac.new(utils.encode(secret), utils.encode(data), hashlib.sha1)
    return utils.decode(base64.b64encode(utils.encode(h.digest())))


def defaultHeaders():
    if not g.userIdHeaderField:
        raise error.ApiError("UserIdHeaderField not set")
    return {
        g.userIdHeaderField: g.userId,
        g.userSignatureHeaderField: signatureV1(g.accessSecret, g.userId),
        g.userSignVersionHeaderField: "v1",
    }


def session(auth=True):
    sess = requests.Session()
    if auth:
        sess.headers.update(defaultHeaders())
    return sess


def request(method, url, *args, **kwargs):
    sess = session(auth=kwargs.get("auth", True))
    func = getattr(sess, method)
    rep = func(url, *args, **kwargs)
    logger.debug(
        f"[{rep.status_code}] {rep.url} - {rep.request.method} - {rep.content}"
    )
    rep.raise_for_status()
    result = rep.json()
    if not result.get("success", True):
        raise error.ApiError(f"Request failed {result}")
    return result


def get(url, *args, **kwargs):
    return request("get", url, *args, **kwargs)


def post(url, *args, **kwargs):
    return request("post", url, *args, **kwargs)


def put(url, *args, **kwargs):
    return request("put", url, *args, **kwargs)


def delete(url, *args, **kwargs):
    return request("delete", url, *args, **kwargs)



# coding=utf-8
from __future__ import absolute_import, print_function

import functools

import tqdm

from suanpan import utils

DEFAULT_PBAR_FORMAT = "{desc}: {n_fmt}/{total_fmt} |{bar}"
DEFAULT_PBAR_CONFIG = {"bar_format": DEFAULT_PBAR_FORMAT}


def parseConfig(config):
    if config is True:
        return utils.merge(DEFAULT_PBAR_CONFIG, {"desc": "Processing"})
    if isinstance(config, str):
        return utils.merge(DEFAULT_PBAR_CONFIG, {"desc": config})
    if config in (False, None):
        return utils.merge(DEFAULT_PBAR_CONFIG, {"disable": True})
    if isinstance(config, dict):
        return utils.merge(DEFAULT_PBAR_CONFIG, config)
    raise Exception(f"Invalid pbar config: bool | str | dict. but {config}")


def runner(pbar, quantity=1):
    def _dec(func):
        @functools.wraps(runner)
        def _runner(*args, **kwargs):
            result = func(*args, **kwargs)
            pbar.update(quantity)
            return result

        return _runner

    return _dec


def getIterableLen(iterable=None, config=None, total=None):
    if config and total is None and iterable is not None:
        iterable = list(iterable)
        total = len(iterable)
    return iterable, total


def one(iterable=None, config=True, total=None, **kwargs):
    iterable, total = getIterableLen(iterable, config=config, total=total)
    params = parseConfig(config)
    params.update(total=total)
    params.update(kwargs)
    return tqdm.tqdm(iterable, **params)


def multi(iterables=None, configs=True, totals=None, **kwargs):
    return Group(iterables=iterables, configs=configs, totals=totals, **kwargs)


class Group(object):
    def __init__(self, iterables=None, configs=True, totals=None, **kwargs):
        lengths = [
            len(i) for i in (iterables, configs, totals) if isinstance(i, (tuple, list))
        ]
        length = min(lengths) if lengths else 0
        if not isinstance(iterables, (tuple, list)):
            iterables = [iterables] * length
        if not isinstance(configs, (tuple, list)):
            configs = [configs] * length
        if not isinstance(totals, (tuple, list)):
            totals = [totals] * length
        self.pbars = [
            one(iterable, config=config, total=total, position=i, **kwargs)
            for i, (iterable, config, total) in enumerate(
                zip(iterables, configs, totals)
            )
        ]

    def add(self, iterable=None, config=True, total=None, **kwargs):
        pbar = one(
            iterable=iterable,
            config=config,
            total=total,
            position=len(self.pbars),
            **kwargs,
        )
        self.pbars.append(pbar)
        return pbar

    def get(self, idx):
        return self.pbars[idx]



import suanpan
from suanpan.app import app
from suanpan.app.arguments import String


@app.output(String(key="outputData1"))
def hello_world(_):
    for i in range(10):
        app.send(str(i))


if __name__ == "__main__":
    suanpan.run(app)



# coding=utf-8
from __future__ import absolute_import, print_function

from contextlib import contextmanager

from suanpan import error

OSS_BUCKET_NAME_CONF = "spark.hadoop.fs.oss.bucketName"


def getStorageHost(spark):
    ossBucketName = spark.sparkContext._conf.get(  # pylint: disable=protected-access
        OSS_BUCKET_NAME_CONF
    )
    if not ossBucketName:
        raise error.SparkError(f"Config {OSS_BUCKET_NAME_CONF} not found")
    return f"oss://{ossBucketName}/"


def getStoragePath(spark, filepath):
    return getStorageHost(spark) + filepath


class File(object):
    def __init__(self, spark, filepath, mode="r"):
        self.spark = spark
        self.filepath = filepath
        self.mode = mode

        self.sc = self.spark.sparkContext
        self.hadoop = self.sc._jvm.org.apache.hadoop  # pylint: disable=protected-access
        self.ioUtils = (
            self.sc._jvm.org.apache.commons.io.IOUtils  # pylint: disable=protected-access
        )
        self.uri = self.sc._jvm.java.net.URI  # pylint: disable=protected-access
        self.fs = self.hadoop.fs.FileSystem
        self.conf = (
            self.sc._jsc.hadoopConfiguration()  # pylint: disable=protected-access
        )
        self.path = self.hadoop.fs.Path(filepath)
        self.fsImpl = self.fs.get(self.uri.create(self.filepath), self.conf)

        self.inputStream = self.fsImpl.open(self.path) if "r" in self.mode else None
        self.overwrite = "a" not in self.mode
        self.outputStream = (
            self.fsImpl.create(self.path, self.overwrite) if "w" in self.mode else None
        )

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.flush()
        self.close()

    def read(self, decode="utf-8"):
        if not self.inputStream:
            raise error.SparkError(f"Error mode {self.mode}")
        return self.ioUtils.toString(self.inputStream, decode)

    def readlines(self, decode="utf-8"):
        if not self.inputStream:
            raise error.SparkError(f"Error mode {self.mode}")
        return self.ioUtils.readLines(self.inputStream, decode)

    def write(self, content, encode="utf-8"):
        if not self.outputStream:
            raise error.SparkError(f"Error mode {self.mode}")
        if not isinstance(content, bytearray):
            content = bytearray(content, encode)
        self.outputStream.write(content)

    def flush(self):
        if self.outputStream:
            self.outputStream.flush()

    def close(self):
        if self.inputStream:
            self.inputStream.close()
        if self.outputStream:
            self.outputStream.close()


class MinioFile(File):
    def __init__(self, spark, filepath, mode="r"):
        filepath = getStoragePath(spark, filepath)
        super(MinioFile, self).__init__(spark, filepath, mode=mode)


@contextmanager
def open(spark, filepath, mode="r", filetype=File):
    file = filetype(spark, filepath, mode=mode)
    yield file
    file.flush()
    file.close()



import os

import suanpan
from suanpan.app import app
from suanpan.app.arguments import Int, Model
from suanpan.model import Model as BaseModel
from suanpan.utils import json


class TestModel(BaseModel):
    def load(self, path):
        self.model = json.load(os.path.join(path, "model.json"))

    def save(self, path):
        json.dump(self.model, os.path.join(path, "model.json"))

    def train(self, data):
        self.model = {"value": data}

    def predict(self, features):
        return self.model["value"]


@app.output(Model(key="outputModel1", type=TestModel))
@app.param(Int(key="param1"))
def test_train(context):
    args = context.args
    args.outputModel1.train(args.param1)
    return args.outputModel1


if __name__ == "__main__":
    suanpan.run(app)



import suanpan
from suanpan.app import app
from suanpan.app.arguments import String

app.sio.disable()


@app.trigger.interval(5)
@app.trigger.output(String(key="outputData1"))
def test2(_):
    return "test"


if __name__ == "__main__":
    suanpan.run(app)



# coding=utf-8
from __future__ import absolute_import, division, print_function

import functools
import math
import os

import oss2
from lostc import collection as lcc
from oss2.models import PartInfo
from oss2.resumable import ResumableDownloadStore, ResumableStore

from suanpan import api, asyncio, g
from suanpan import path as spath
from suanpan import runtime
from suanpan.log import logger
from suanpan.storage import base
from suanpan.utils import pbar as spbar


class Storage(base.Storage):
    LARGE_FILE_SIZE = 1024 * 1024 * 1024  # 1GB
    PREFERRED_SIZE = 500 * 1024 * 1024  # 500MB

    def __init__(
        self,
        ossAccessId=None,
        ossAccessKey=None,
        ossBucketName="suanpan",
        ossEndpoint="http://oss-cn-beijing.aliyuncs.com",
        ossDelimiter="/",
        ossTempStore=base.Storage.DEFAULT_TEMP_DATA_STORE,
        ossGlobalStore=base.Storage.DEFAULT_GLOBAL_DATA_STORE,
        ossDownloadNumThreads=1,
        ossDownloadStoreName=".py-oss-download",
        ossUploadNumThreads=1,
        ossUploadStoreName=".py-oss-upload",
        **kwargs,
    ):
        super(Storage, self).__init__(
            delimiter=ossDelimiter,
            tempStore=ossTempStore,
            globalStore=ossGlobalStore,
            **kwargs,
        )

        self.bucketName = ossBucketName
        self.endpoint = ossEndpoint
        self.refreshAccessKey(accessId=ossAccessId, accessKey=ossAccessKey)

        self.downloadNumThreads = ossDownloadNumThreads
        self.downloadStoreName = ossDownloadStoreName
        self.downloadStore = ResumableDownloadStore(
            self.tempStore, self.downloadStoreName
        )

        self.uploadNumThreads = ossUploadNumThreads
        self.uploadStoreName = ossUploadStoreName
        self.uploadStore = ResumableStore(self.tempStore, self.uploadStoreName)

        self._removeOssLogger()

    def autoRefreshToken(self, func):
        @functools.wraps(func)
        def _dec(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except oss2.exceptions.OssError as e:
                if e.status != 403:
                    raise e
                logger.warning("Oss access denied, refreshing access key.")
                self.refreshAccessKey()
                return func(*args, **kwargs)

        return _dec

    def autoRefreshBucket(self, bucket):
        try:
            bucket.object_exists(f"studio/{g.userId}/permissions")
            return bucket
        except oss2.exceptions.OssError as e:
            if e.status != 403:
                raise e
            logger.warning("Oss access denied, refreshing access key.")
            self.refreshAccessKey()
            bucket = self.getBucketByName(bucket.bucket_name)
            return bucket

    def refreshAccessKey(self, accessId=None, accessKey=None):
        if accessId and accessKey:
            self.accessId = accessId
            self.accessKey = accessKey
            self.auth = oss2.Auth(self.accessId, self.accessKey)
        else:
            data = api.oss.getToken()
            self.accessId = data["Credentials"]["AccessKeyId"]
            self.accessKey = data["Credentials"]["AccessKeySecret"]
            self.securityToken = data["Credentials"]["SecurityToken"]
            self.auth = oss2.StsAuth(self.accessId, self.accessKey, self.securityToken)

        self.bucket = self.getBucket(self.bucketName)
        return self.accessId, self.accessKey

    def _removeOssLogger(self):
        ossLogger = getattr(oss2, "logger", None)
        if ossLogger:
            self._removeLoggerHandlers(ossLogger)

    def _removeLoggerHandlers(self, _logger):
        for handler in _logger.handlers[:]:
            _logger.removeHandler(handler)
        return _logger

    @runtime.retry(stop_max_attempt_number=3)
    def download(self, name, path, bucketOrBucketName=None, ignores=None):
        bucket = self.getBucket(bucketOrBucketName)
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        downloadFunction = (
            self.downloadFile
            if self.isFile(name, bucketOrBucketName=bucket)
            else self.downloadFolder
        )
        return downloadFunction(name, path, bucketOrBucketName=bucket, ignores=ignores)

    def downloadFolder(
        self,
        folderName,
        folderPath,
        delimiter=None,
        bucketOrBucketName=None,
        ignores=None,
    ):
        bucket = self.getBucket(bucketOrBucketName)
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        delimiter = delimiter or self.delimiter
        storagePath = self.storageUrl(folderName, bucketOrBucketName=bucket)

        if folderPath in ignores:
            logger.debug(f"Ignore downloading folder: {folderPath} -> {storagePath}")
            return folderPath

        downloads = {
            file: self.localPathJoin(
                folderPath, self.storageRelativePath(file, folderName)
            )
            for _, _, files in self.walk(
                folderName, delimiter=delimiter, bucketOrBucketName=bucket
            )
            for file in files
        }

        logger.info(f"Downloading folder: {storagePath} -> {folderPath}")
        # Download from oss
        _run = functools.partial(
            self.downloadFile, bucketOrBucketName=bucket, ignores=ignores, quiet=True
        )
        asyncio.starmap(_run, downloads.items(), pbar="Downloading")
        # Remove ignores
        self.removeIgnores(folderPath, ignores=ignores)
        # Remove rest files and folders
        files = (
            os.path.join(root, file)
            for root, _, files in os.walk(folderPath)
            for file in files
        )
        restFiles = [file for file in files if file not in downloads.values()]
        asyncio.map(
            spath.remove, restFiles, pbar="Removing Rest Files" if restFiles else False,
        )
        spath.removeEmptyFolders(folderPath)
        logger.debug(f"Removed empty folders in: {folderPath}")
        # End
        logger.debug(f"Downloaded folder: {storagePath} -> {folderPath}")
        return folderPath

    def downloadFile(
        self, objectName, filePath, bucketOrBucketName=None, ignores=None, quiet=False
    ):
        bucket = self.getBucket(bucketOrBucketName)
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        storagePath = self.storageUrl(objectName, bucketOrBucketName=bucket)
        fileSize = self.getStorageSize(objectName, bucketOrBucketName=bucket)

        if not quiet:
            logger.info(f"Downloading file: {storagePath} -> {filePath}")

        with spbar.one(total=fileSize, config=not quiet) as pbar:
            if filePath in ignores:
                pbar.update(fileSize)
                pbar.set_description("Ignored")
                return filePath

            fileMd5 = self.getLocalMd5(filePath)
            objectMd5 = self.getStorageMd5(objectName, bucketOrBucketName=bucket)
            if self.checkMd5(fileMd5, objectMd5):
                pbar.update(fileSize)
                pbar.set_description("Existed")
                return filePath

            def _percentage(consumed_bytes, total_bytes):
                if total_bytes:
                    pbar.update(consumed_bytes - pbar.n)
                    pbar.set_description("Downloading")

            spath.safeMkdirsForFile(filePath)
            self.autoRefreshToken(oss2.resumable_download)(
                bucket,
                objectName,
                filePath,
                num_threads=self.downloadNumThreads,
                store=self.downloadStore,
                progress_callback=_percentage,
            )

            pbar.set_description("Downloaded")

            return filePath

    @runtime.retry(stop_max_attempt_number=3)
    def upload(self, name, path, bucketOrBucketName=None, ignores=None):
        bucket = self.getBucket(bucketOrBucketName)
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        uploadFunction = self.uploadFolder if os.path.isdir(path) else self.uploadFile
        return uploadFunction(name, path, bucketOrBucketName=bucket, ignores=ignores)

    def uploadFolder(
        self, folderName, folderPath, bucketOrBucketName=None, ignores=None,
    ):
        bucket = self.getBucket(bucketOrBucketName)
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        storagePath = self.storageUrl(folderName, bucketOrBucketName=bucket)

        if folderName in ignores:
            logger.debug(f"Ignore uploading folder: {folderName} -> {storagePath}")
            return folderPath

        filePaths = (
            os.path.join(root, file)
            for root, _, files in os.walk(folderPath)
            for file in files
        )
        uploads = {
            filePath: self.storagePathJoin(
                folderName, self.localRelativePath(filePath, folderPath)
            )
            for filePath in filePaths
        }

        if not uploads:
            logger.warning(f"Uploading empty folder: {folderPath}")
        else:
            logger.info(f"Uploading folder: {folderPath} -> {storagePath}")
            # Upload files to oss
            uploadItems = [
                (objectName, filePath) for filePath, objectName in uploads.items()
            ]
            _run = functools.partial(
                self.uploadFile, bucketOrBucketName=bucket, ignores=ignores, quiet=True
            )
            asyncio.starmap(_run, uploadItems, pbar="Uploading")

        # Remove rest files
        localFiles = set(
            self.localRelativePath(filePath, folderPath) for filePath in uploads.keys()
        )
        remoteFiles = set(
            self.storageRelativePath(objectName, folderName)
            for _, _, files in self.walk(folderName)
            for objectName in files
        )
        restFiles = [
            self.storagePathJoin(folderName, file) for file in remoteFiles - localFiles
        ]
        _run = functools.partial(self.remove, bucketOrBucketName=bucket, quiet=True)
        asyncio.map(
            _run, restFiles, pbar="Removing Rest Files" if restFiles else False,
        )
        # End
        logger.debug(f"Uploaded folder: {folderPath} -> {storagePath}")
        return folderPath

    def uploadFile(
        self, objectName, filePath, bucketOrBucketName=None, ignores=None, quiet=False
    ):
        bucket = self.getBucket(bucketOrBucketName)
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        storagePath = self.storageUrl(objectName, bucketOrBucketName=bucket)
        fileSize = self.getLocalSize(filePath)

        if not quiet:
            logger.info(f"Uploading file: {filePath} -> {storagePath}")

        with spbar.one(total=fileSize, config=not quiet) as pbar:
            if filePath in ignores:
                pbar.update(fileSize)
                pbar.set_description("Ignored")
                return filePath

            fileMd5 = self.getLocalMd5(filePath)
            objectMd5 = self.getStorageMd5(objectName, bucketOrBucketName=bucket)
            if self.checkMd5(fileMd5, objectMd5):
                pbar.update(fileSize)
                pbar.set_description("Existed")
                return filePath

            def _percentage(consumed_bytes, total_bytes):
                if total_bytes:
                    pbar.update(consumed_bytes - pbar.n)
                    pbar.set_description("Uploading")

            self.autoRefreshToken(oss2.resumable_upload)(
                bucket,
                objectName,
                filePath,
                num_threads=self.uploadNumThreads,
                store=self.uploadStore,
                progress_callback=_percentage,
                headers={self.CONTENT_MD5: fileMd5},
            )

            pbar.set_description("Uploaded")

            return filePath

    @runtime.retry(stop_max_attempt_number=3)
    def copy(self, name, dist, bucketOrBucketName=None):
        bucket = self.getBucket(bucketOrBucketName)
        copyFunction = (
            self.copyFile
            if self.isFile(name, bucketOrBucketName=bucket)
            else self.copyFolder
        )
        return copyFunction(name, dist, bucketOrBucketName=bucket)

    def copyFolder(
        self, folderName, distName, bucketOrBucketName=None, delimiter=None,
    ):
        bucket = self.getBucket(bucketOrBucketName)
        delimiter = delimiter or self.delimiter
        folderName = self.completePath(folderName)
        distName = self.completePath(distName)
        logger.info(f"Copying folder: {folderName} -> {distName}")
        copyItems = [
            (file, file.replace(folderName, distName))
            for _, _, files in self.walk(
                folderName, delimiter=delimiter, bucketOrBucketName=bucket
            )
            for file in files
        ]
        _run = functools.partial(self.copyFile, bucketOrBucketName=bucket, quiet=True)
        asyncio.starmap(_run, copyItems, pbar="Copying")

    def copyFile(self, objectName, distName, bucketOrBucketName=None, quiet=False):
        bucket = self.getBucket(bucketOrBucketName)
        fileSize = self.getStorageSize(objectName, bucketOrBucketName=bucket)
        copyFunction = (
            self.copyLargeFile
            if fileSize >= self.LARGE_FILE_SIZE
            else self.copySmallFile
        )
        return copyFunction(
            objectName, distName, fileSize, bucketOrBucketName=bucket, quiet=quiet
        )

    def copySmallFile(
        self, objectName, distName, size, bucketOrBucketName=None, quiet=False
    ):
        bucket = self.getBucket(bucketOrBucketName)

        if not quiet:
            _src = self.storageUrl(objectName, bucketOrBucketName=bucket)
            _target = self.storageUrl(distName, bucketOrBucketName=bucket)
            logger.info(f"Copying file: {_src} -> {_target}")

        with spbar.one(total=size, config=not quiet) as pbar:
            objectMd5 = self.getStorageMd5(objectName, bucketOrBucketName=bucket)
            distMd5 = self.getStorageMd5(distName, bucketOrBucketName=bucket)
            if self.checkMd5(objectMd5, distMd5):
                pbar.update(size)
                pbar.set_description("Existed")
                return distName

            bucket.copy_object(bucket.bucket_name, objectName, distName)
            pbar.update(size)
            return distName

    def copyLargeFile(
        self, objectName, distName, size, bucketOrBucketName=None, quiet=False
    ):
        bucket = self.getBucket(bucketOrBucketName)

        if not quiet:
            _src = self.storageUrl(objectName, bucketOrBucketName=bucket)
            _target = self.storageUrl(distName, bucketOrBucketName=bucket)
            logger.info(f"Copying file: {_src} -> {_target}")

        with spbar.one(total=size, config=not quiet) as pbar:
            objectMd5 = self.getStorageMd5(objectName, bucketOrBucketName=bucket)
            distMd5 = self.getStorageMd5(objectName, bucketOrBucketName=bucket)
            if self.checkMd5(objectMd5, distMd5):
                pbar.update(size)
                pbar.set_description("Existed")
                return distName

            partSize = self.autoRefreshToken(oss2.determine_part_size)(
                size, preferred_size=self.PREFERRED_SIZE
            )
            uploadId = bucket.init_multipart_upload(distName).upload_id
            parts = math.ceil(size / partSize)
            parts = (
                (i + 1, i * partSize, min((i + 1) * partSize, size))
                for i in range(parts)
            )

            def _copy(part):
                partNumber, byteRange = part[0], part[-2:]
                result = bucket.upload_part_copy(
                    bucket.bucket_name,
                    objectName,
                    byteRange,
                    distName,
                    uploadId,
                    partNumber,
                )
                pbar.update(byteRange[1] - byteRange[0])
                return PartInfo(partNumber, result.etag)

            parts = [_copy(part) for part in parts]
            bucket.complete_multipart_upload(distName, uploadId, parts)
            return distName

    @runtime.retry(stop_max_attempt_number=3)
    def remove(
        self, objectName, delimiter=None, bucketOrBucketName=None, quiet=False,
    ):
        delimiter = delimiter or self.delimiter
        bucket = self.getBucket(bucketOrBucketName)
        removeFunc = (
            self.removeFile
            if self.isFile(objectName, bucketOrBucketName=bucket)
            else self.removeFolder
        )
        return removeFunc(
            objectName, delimiter=delimiter, bucketOrBucketName=bucket, quiet=quiet
        )

    def removeFolder(
        self, folderName, delimiter=None, bucketOrBucketName=None, quiet=False,
    ):
        delimiter = delimiter or self.delimiter
        bucket = self.getBucket(bucketOrBucketName)
        folderName = self.completePath(folderName)
        removes = [
            objectName for _, _, files in self.walk(folderName) for objectName in files
        ]
        _run = functools.partial(
            self.remove, delimiter=delimiter, bucketOrBucketName=bucket, quiet=True
        )
        asyncio.map(
            _run, removes, pbar="Removing" if removes and not quiet else False,
        )
        return folderName

    def removeFile(
        self, fileName, delimiter=None, bucketOrBucketName=None, quiet=False
    ):  # pylint: disable=unused-argument
        bucket = self.getBucket(bucketOrBucketName)
        bucket.delete_object(fileName)
        if not quiet:
            storagePath = self.storageUrl(fileName, bucketOrBucketName=bucket)
            logger.info(f"Removed file: {storagePath}")
        return fileName

    def walk(self, folderName, delimiter=None, bucketOrBucketName=None):
        bucket = self.getBucket(bucketOrBucketName)
        delimiter = delimiter or self.delimiter
        root = self.completePath(folderName, delimiter=delimiter)
        objects = (
            obj
            for obj in runtime.saferun(self._listAll, default=iter(()))(
                folderName, delimiter=delimiter, bucketOrBucketName=bucket
            )
            if obj.key != root
        )
        folders, files = lcc.divide(objects, lambda obj: obj.is_prefix())
        folders, files = self._getObjectNames(folders), self._getObjectNames(files)
        yield root, folders, files
        for folder in folders:
            for item in self.walk(
                folder, delimiter=delimiter, bucketOrBucketName=bucket
            ):
                yield item

    def _listAll(self, folderName, delimiter=None, bucketOrBucketName=None):
        bucket = self.getBucket(bucketOrBucketName)
        delimiter = delimiter or self.delimiter
        prefix = (
            folderName if folderName.endswith(delimiter) else folderName + delimiter
        )
        return (
            obj
            for obj in self.autoRefreshToken(oss2.ObjectIterator)(
                delimiter=delimiter, prefix=prefix, bucket=bucket
            )
        )

    def listAll(self, folderName, delimiter=None, bucketOrBucketName=None):
        return (
            obj.key
            for obj in self._listAll(
                folderName, delimiter=delimiter, bucketOrBucketName=bucketOrBucketName
            )
        )

    def listFolders(self, folderName, delimiter=None, bucketOrBucketName=None):
        return (
            obj.key
            for obj in self._listAll(
                folderName, delimiter=delimiter, bucketOrBucketName=bucketOrBucketName
            )
            if obj.is_prefix()
        )

    def listFiles(self, folderName, delimiter=None, bucketOrBucketName=None):
        return (
            obj.key
            for obj in self._listAll(
                folderName, delimiter=delimiter, bucketOrBucketName=bucketOrBucketName
            )
            if not obj.is_prefix()
        )

    def isFolder(self, folderName, bucketOrBucketName=None):
        bucket = self.getBucket(bucketOrBucketName)
        return next(self.listAll(folderName, bucketOrBucketName=bucket), None)

    def isFile(self, objectName, bucketOrBucketName=None):
        bucket = self.getBucket(bucketOrBucketName)
        return bucket.object_exists(objectName)

    def getStorageMd5(self, objectName, bucketOrBucketName=None):
        bucket = self.getBucket(bucketOrBucketName)
        try:
            return bucket.head_object(objectName).headers.get(self.CONTENT_MD5)
        except oss2.exceptions.NotFound:
            return None

    def getStorageSize(self, objectName, bucketOrBucketName=None):
        bucket = self.getBucket(bucketOrBucketName)
        return bucket.head_object(objectName).content_length

    def storageUrl(self, path, bucketOrBucketName=None):
        bucket = self.getBucket(bucketOrBucketName)
        return "oss:///" + self.storagePathJoin(bucket.bucket_name, path)

    def getBucket(self, bucketOrBucketName):
        return self.autoRefreshBucket(
            bucketOrBucketName
            if isinstance(bucketOrBucketName, oss2.Bucket)
            else self.getBucketByName(bucketOrBucketName)
        )

    def getBucketByName(self, bucketName=None):
        return (
            oss2.Bucket(self.auth, self.endpoint, bucketName)
            if bucketName
            else self.bucket
        )

    def _getObjectNames(self, objects):
        return (
            [obj.key for obj in objects]
            if isinstance(objects, (tuple, list))
            else objects.key
        )



# coding=utf-8
from __future__ import absolute_import, print_function
from __suanpan__ import asyncio  # pylint: disable=unused-import

import gevent  # pylint: disable=wrong-import-order

from suanpan.utils import pbar as spbar


def imap(func, iterable, timeout=None, count=None, pbar=None):
    return iwait(
        [gevent.spawn(func, i) for i in iterable],
        timeout=timeout,
        count=count,
        pbar=pbar,
    )


def map(func, iterable, timeout=None, count=None, pbar=None):
    return list(imap(func, iterable, timeout=timeout, count=count, pbar=pbar))


def istarmap(func, iterable, timeout=None, count=None, pbar=None):
    return iwait(
        [gevent.spawn(func, *i) for i in iterable],
        timeout=timeout,
        count=count,
        pbar=pbar,
    )


def starmap(func, iterable, timeout=None, count=None, pbar=None):
    return list(istarmap(func, iterable, timeout=timeout, count=count, pbar=pbar))


def run(funcs, *args, **kwargs):
    return [gevent.spawn(func, *args, **kwargs) for func in funcs]


def wait(objects, timeout=None, count=None, pbar=None):
    return list(iwait(objects, timeout=timeout, count=count, pbar=pbar))


def iwait(iterable, timeout=None, count=None, pbar=None):
    iterable, total = spbar.getIterableLen(iterable, config=pbar, total=count)
    iterable = _iwait(iterable, timeout=timeout, count=count)
    return spbar.one(iterable, config=pbar, total=total)


def _iwait(objects, timeout=None, count=None):
    for obj in gevent.iwait(objects, timeout=timeout, count=count):
        if getattr(obj, "exception", None) is not None:
            if hasattr(obj, "_raise_exception"):
                obj._raise_exception()  # pylint: disable=protected-access
            else:
                raise obj.exception
        yield obj


def sleep(seconds=0, ref=True):
    return gevent.sleep(seconds=seconds, ref=ref)


def switch():
    return sleep(0)


def kill(iterable, exception=gevent.GreenletExit, block=True, timeout=None):
    iterable = iterable if isinstance(iterable, (list, tuple)) else [iterable]
    return gevent.killall(iterable, exception=exception, block=block, timeout=timeout)


def current():
    return gevent.getcurrent()



# coding=utf-8
from __future__ import absolute_import, print_function

import abc

from suanpan.mstorage.vars import Variables
from suanpan.objects import HasName


class MStorage(HasName):
    @abc.abstractmethod
    def get(self, name, *args, **kwargs):
        pass

    @abc.abstractmethod
    def set(self, name, value, *args, **kwargs):
        pass

    @abc.abstractmethod
    def mget(self, name, *args, **kwargs):
        pass

    @abc.abstractmethod
    def mset(self, name, mapping, *args, **kwargs):
        pass

    @property
    def vars(self):
        return Variables(self)



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.app import app
from suanpan.app.modules.base import Module

module = Module()


@module.on("connect")
def notifyParamsConfigs(context):
    app.sio.emit("params.configs.notify", getParamsConfigs(context))
    app.sio.emit("params.notify", getParams(context))


@module.on("params.configs.get")
def getParamsConfigs(_):
    configs = app.sio._loop.options.params.configs  # pylint: disable=protected-access
    stream = app._stream  # pylint: disable=protected-access
    return {
        "title": configs.title or stream.title,
        "params": configs.params or [arg.paramConfigs for arg in stream.ARGUMENTS],
    }


@module.on("params.get")
def getParams(_):
    stream = app._stream  # pylint: disable=protected-access
    return stream.getParams()


@module.on("params.update")
def updateParams(context):
    stream = app._stream  # pylint: disable=protected-access
    loop = app._loop  # pylint: disable=protected-access
    params = stream.updateParams(context.args)
    loop.retryLastMessage()
    params = stream.saveParams(params)
    return params



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.error import base


class UnknownBaseServicesError(base.Error):
    MESSAGE = "Suanpan Unknown Base Services Error: {}"



# coding=utf-8
from __future__ import absolute_import, print_function

import datetime
import io
import traceback
import uuid

from suanpan import path, runtime
from suanpan.log import logger
from suanpan.stream import handlers
from suanpan.stream.interfaces import HasHandlers


class Event(object):
    TEMPLATE = "__{keyword}_::{event}::{data}__"

    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self._kwargs.setdefault("keyword", "akuma")

    def __call__(self, event, *args, **kwargs):
        kwargs.update(event=event)
        return self.format(*args, **kwargs)

    def format(self, *args, **kwargs):
        return self.TEMPLATE.format(*self._args, *args, **self._kwargs, **kwargs)

    def new(self):
        return self.format(event="new", data="id")

    def resume(self, id=""):
        return self.format(event="resume", data=id)

    def data(self, id):
        return self.format(event="data", data=id)

    def more(self, id):
        return self.format(event="more", data=id)

    def stop(self):
        return self.format(event="stop", data="")

    def end(self, id):
        return self.format(event="end", data=id)


class Stream(object):
    def __init__(self, id):
        self.record = {
            "id": id,
            "event": None,
            "size": 0,
            "total": 0,
            "expire": self._addExipre(new=True),
            "data": {},
        }
        self.buffer = io.BytesIO()

    def _addExipre(self, expire=None, new=False):
        expire = expire or self.now
        if new:
            return expire + datetime.timedelta(minutes=5)
        if (self.now - expire).seconds <= 60:
            return expire + datetime.timedelta(milliseconds=60)
        return expire

    @property
    def now(self):
        return datetime.datetime.now()

    @property
    def sid(self):
        return self.record["sid"]

    @property
    def event(self):
        return self.record["event"]

    @property
    def size(self):
        return self.record["size"]

    @property
    def total(self):
        return self.record["total"]

    @property
    def dirty(self):
        return self.record["dirty"]

    @property
    def expire(self):
        return self.record["expire"]

    @property
    def data(self):
        return self.record["data"]

    @property
    def expired(self):
        return self.now > self.expire

    @property
    def done(self):
        return self.total >= self.size

    def update(self, *args, **kwargs):
        self.record.update(*args, **kwargs)
        return self

    def read(self):
        self.buffer.seek(0)
        data = self.buffer.read()
        self.buffer.truncate(0)
        self.buffer.seek(0)
        return data

    def write(self, chunks):
        if not chunks:
            return 0

        total = self.buffer.write(chunks)
        self.update(total=self.total + total, expire=self._addExipre(self.expire))
        return total

    def pipe(self, file):
        path.mkdirs(file, parent=True)
        chunks = self.read()
        if self.total - len(chunks) <= 0:
            path.remove(file)
        with open(file, "ab") as f:
            return f.write(chunks)


class Cleaner(object):
    def __init__(self, delta, handler):
        self.delta = delta
        self.handler = handler
        self.expire = self.now + delta

    @property
    def now(self):
        return datetime.datetime.now()

    @property
    def expired(self):
        return self.now > self.expire

    def clean(self, *args, **kwargs):
        if self.expired:
            self.handler(*args, **kwargs)


class StreamServer(HasHandlers):
    DEFAULT_HANDLER_CLASS = handlers.SIOStreamHandler

    def __init__(self, stream, sio):
        super(StreamServer, self).__init__()
        self.stream = stream
        self.sio = sio
        self.streams = {}
        self.event = Event()
        self.cleaner = Cleaner(datetime.timedelta(seconds=10), self._clean)
        self.init()

    def init(self):
        self.sio.on(self.event.new(), self.onNew)
        self.sio.on(self.event.resume(), self.onResume)

    @property
    def now(self):
        return datetime.datetime.now()

    def _addTime(self, date, new=False):
        if new:
            return date + datetime.timedelta(minutes=5)
        if (self.now - date).seconds <= 60:
            return date + datetime.timedelta(milliseconds=60)
        return date

    def _genid(self):
        id = uuid.uuid4().hex
        while id in self.streams:
            id = uuid.uuid4().hex
        return id

    def _createNew(self, id):
        self._listener(id)
        return id

    def _listener(self, id):
        self.sio.on(self.event.data(id), lambda sid, data: self.onData(sid, id, data))

    def _done(self, id, *args, **kwargs):
        self.end(id, *args, **kwargs)
        self.streams.pop(id)

    def _clean(self):
        for id, stream in self.streams.items():
            if stream.expired:
                self.streams.pop(id)

    def _handle(self, key, handler):
        def _dec(sid, data, stream):
            title = f"{sid} - sio:{key}"
            _run = runtime.costrun(title)(handler.run)
            try:
                return _run(self.stream, data, sid, stream)
            except Exception:  # pylint: disable=broad-except
                logger.error(traceback.format_exc())

        return _dec

    def getHandler(self, key):
        if not self.hasHandler(key):
            super(StreamServer, self).setHandler(key, self.DEFAULT_HANDLER_CLASS())
        return super(StreamServer, self).getHandler(key)

    def setHandler(self, key, handler):
        return self.getHandler(key).use(handler)

    def clean(self, *args, **kwargs):
        self.cleaner.clean(*args, **kwargs)

    def onNew(self, sid):  # pylint: disable=unused-argument
        return self._createNew(self._genid())

    def onData(self, sid, id, data):
        self.clean()

        stream = self.streams.get(id)
        if not stream:
            stream = Stream(id).update(event=data["event"], **data["info"])
            self.streams[id] = stream
        stream.write(data["chunk"])

        payload = None
        if self.hasHandler(stream.event):
            payload = self._handle(stream.event, self.getHandler(stream.event))(
                sid, stream.data, stream
            )

        if stream.done:
            self._done(id, {"total": stream.total, "payload": payload})
        else:
            self.more(id, stream.total)

    def onStop(self, sid, id):  # pylint: disable=unused-argument
        if id in self.streams:
            self._done(id)

    def onResume(self, sid, id):  # pylint: disable=unused-argument
        stream = self.streams.get(id)
        if stream:
            self.resume(id, stream.total)
        else:
            self._createNew(id)
            self.resume(id)

    def more(self, id, *args, **kwargs):
        self.sio.emit(self.event.more(id), *args, **kwargs)

    def resume(self, id, *args, **kwargs):
        self.sio.emit(self.event.resume(id), *args, **kwargs)

    def end(self, id, *args, **kwargs):
        self.sio.emit(self.event.end(id), *args, **kwargs)



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.app.base import BaseApp
from suanpan.spark import SparkComponent as Handler


class SparkApp(BaseApp):
    def __init__(self, *args, **kwargs):
        super(SparkApp, self).__init__(*args, **kwargs)
        self.handler = Handler()

    def __call__(self, funcOrApp):
        if not isinstance(funcOrApp, BaseApp):
            self.handler.use(funcOrApp)
        return self

    @property
    def name(self):
        return self.handler.name

    def start(self, *args, **kwargs):
        super(SparkApp, self).start(*args, **kwargs)
        return self.handler.start(*args, **kwargs)

    def input(self, argument):
        self.handler.input(argument)
        return self

    def output(self, argument):
        self.handler.output(argument)
        return self

    def param(self, argument):
        self.handler.param(argument)
        return self

    def column(self, argument):
        self.handler.column(argument)
        return self

    def beforeInit(self, hook):
        self.handler.addBeforeInitHooks(hook)
        return hook

    def afterInit(self, hook):
        self.handler.addAfterInitHooks(hook)
        return hook

    def beforeCall(self, hook):
        self.handler.addBeforeCallHooks(hook)
        return hook

    def afterCall(self, hook):
        self.handler.addAfterCallHooks(hook)
        return hook

    def beforeExit(self, hook):
        self.handler.addBeforeExitHooks(hook)
        return hook

    def load(self, *args, **kwargs):
        return self.handler.load(*args, **kwargs)

    def save(self, *args, **kwargs):
        return self.handler.save(*args, **kwargs)

    def send(self, *args, **kwargs):
        return self.handler.save(*args, **kwargs)



# coding=utf-8
from __future__ import absolute_import, print_function

from addict import Dict


class HasName(object):
    @property
    def name(self):
        return self.__class__.__name__


class Context(Dict, HasName):
    @classmethod
    def froms(cls, *args, **kwargs):
        context = cls()
        context.update(item for arg in args for item in arg.items())
        context.update(kwargs)
        return context

    def __getitem__(self, name):
        value = super(Context, self).__getitem__(name)
        if isinstance(value, property):
            value = value.fget(self)
            self.__setitem__(name, value)
        return value



# coding=utf-8
from __future__ import absolute_import, print_function

import abc
import time

from suanpan import objects


class Loop(objects.HasName):
    __metaclass__ = abc.ABCMeta

    def __call__(self):
        return self.loop()

    @abc.abstractmethod
    def loop(self):
        pass


class Interval(Loop):
    def __init__(self, seconds, pre=False):
        self.seconds = seconds
        self.pre = pre

    def loop(self):
        while True:
            if self.pre:
                time.sleep(self.seconds)
            yield
            if not self.pre:
                time.sleep(self.seconds)

    def set(self, seconds):
        self.seconds = seconds
        return self


class IntervalIndex(Interval):
    def loop(self):
        for i, _ in enumerate(super(IntervalIndex, self).loop()):
            yield i



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.error import base


class AppError(base.Error):
    MESSAGE = "Suanpan APP Error: {}"


class AppModuleError(AppError):
    MESSAGE = "Suanpan APP Module Error: {}"



# coding=utf-8
from __future__ import absolute_import, print_function

import math
import operator
from functools import reduce

import numpy as np


def toND(data, n):
    if len(data.shape) < n - 1:
        raise Exception(f"Data shape should more than {n-1}D: {data.shape}")
    keeps = data.shape[1 - n :]
    layers = reduce(operator.mul, data.shape[: 1 - n], 1)
    return data.reshape((layers,) + keeps)


def to2D(data):
    return toND(data, 2)


def to3D(data):
    return toND(data, 3)


def to4D(data):
    return toND(data, 4)


def flatAsImage(data):
    data = to4D(data) if len(data.shape) > 3 and data.shape[-1] == 3 else to3D(data)
    layers, height, width = data.shape[:3]
    columnNumber = int(math.ceil(math.sqrt(layers)))
    rowNumber = layers // columnNumber + 1 if columnNumber > 1 else 1
    extraLayers = rowNumber * columnNumber - layers
    extraShape = [extraLayers, height, width]
    extraShape[3:] = data.shape[3:]
    extraData = np.zeros(extraShape)
    fullData = np.concatenate([data, extraData])
    rows = [
        np.concatenate(fullData[i * columnNumber : (i + 1) * columnNumber], axis=1)
        for i in range(rowNumber)
    ]
    return np.concatenate(rows)



# coding=utf-8
from __future__ import absolute_import, print_function

import collections

from suanpan.log import logger


class Module(object):
    def __init__(self, enabled=True):
        super(Module, self).__init__()
        self._enabled = enabled
        self.handlers = collections.defaultdict(list)
        self.streamHandlers = collections.defaultdict(list)
        self.functions = {}
        self.bootstraps = {}

    def enable(self):
        self._enabled = True
        return self

    def disable(self):
        self._enabled = False
        return self

    @property
    def enabled(self):
        return self._enabled

    def on(self, event):
        def _dec(func):
            self.handlers[event].append(func)
            return func

        return _dec

    def stream(self, event):
        def _dec(func):
            self.streamHandlers[event].append(func)
            return func

        return _dec

    def fn(self, fn):
        self.functions[fn.__name__] = fn
        return fn

    def __getattr__(self, key):
        if key in self.functions:
            return self.functions[key]
        return self.__getattribute__(key)

    def _bootstrap(self):
        for _, bootstrap in self.bootstraps.items():
            bootstrap()

    def _onBootstrap(self, fn):
        self.bootstraps[fn.__name__] = fn
        return fn

    def bootstrap(self, fn=None):
        if fn:
            return self._onBootstrap(fn)
        return self._bootstrap()


class Modules(object):
    def __init__(self):
        super(Modules, self).__init__()
        self.modules = {}

    def __getattr__(self, key):
        if key in self.modules:
            return self.modules[key]
        return self.__getattribute__(key)

    def register(self, name, module, enable=True):
        module = module.enable() if enable else module.disable()
        self.modules[name] = module
        return module

    @property
    def handlers(self):
        _handlers = collections.defaultdict(list)
        for module in self.modules.values():
            if module.enabled:
                for event, fns in module.handlers.items():
                    _handlers[event].extend(fns)
        return _handlers

    @property
    def streamHandlers(self):
        _handlers = collections.defaultdict(list)
        for module in self.modules.values():
            if module.enabled:
                for event, fns in module.streamHandlers.items():
                    _handlers[event].extend(fns)
        return _handlers

    def enable(self, name):
        return self.modules[name].enable()

    def disable(self, name):
        return self.modules[name].disable()

    def info(self):
        for name, module in self.modules.items():
            enabledStr = "Enabled" if module.enabled else "Disabled"
            logger.debug(f"App Module {name}: {enabledStr}")

    def bootstrap(self):
        for _, module in self.modules.items():
            if module.enabled:
                module.bootstrap()
